UNIT uWortListe;
(* ******************************************************************** *)
(*                                                                      *)
(* K L A S S E   : TWortListe                                           *)
(* -------------------------------------------------------------------- *)
(* Version       : 1.31                                                  *)
(* Autor         : S. Spolwig, OSZ-Handel I, 10997 Berlin               *)
(*                                                                      *)
(* Aufgabe       : Allgemeine statische Liste zur Verwaltung beliebiger *)
(*                 Objekte.                                             *)
(*                 Zum Bewegen in der Liste ist intern eine Listenmarke *)
(*                 mitgefuehrt, die die jeweils aktuelle Listenposition *)
(*                 bezeichnet. Alle Operationen werden auf der aktuellen*)
(*                 Position ausgefuehrt.                                *)
(*                 (SLISTE entspricht SLISTEN in Turbo-Pascal.)         *)
(*                                                                      *)
(* Compiler      : MSDOS - Turbo-Pascal 7.0 - Delphi 3.0                *)
(* Aenderung     : V. 1.0   - 20-AUG-97                                 *)
(*                    1.02  - 03-SEP-98   Load, Store                   *)
(*                    1.03  - 06-Okt-98   SearchElement                 *)
(*                    1.2   - 01-JAN-99   als Stringliste               *)
(*                                        GetSuchPos,                   *)
(*                    1.3   - 04-JAN-99   GetSuchPosBin; RekursivDurchs *)
(*                                        Sort                          *)
(*                            24-FEB-00   an SListe angepasst           *)
(*                    --------------------------------------------      *)
(*                            16-MAR-00   Stringliste zum sortieren     *)
(*                    1.31     22-FEB-02   TwortListe erbt StrLIste      *)
(* ******************************************************************** *)

INTERFACE
(* ==================================================================== *)
uses
  uWortGenerator,  math, dialogs, sysutils, ComCtrls,
  uStrListe;

type
  TWortListe  = class(TStrListe)
   private
    SingleMode  : boolean;  // ob jedes El. nur einmal drin ist
    ElemLaenge  : Cardinal;
    ElemCase    : boolean;
   public
    procedure PegelAnzeigen(pb: TProgressbar);
    procedure PegelStart(max : Cardinal);
    procedure SetLen (L: Cardinal);           virtual;

    procedure NeuesWortMachen;               virtual;
    procedure SetElemLen (l : Cardinal);      virtual;
    procedure SetElemCase(c : boolean);      virtual;
    procedure SetSingleMode (m : boolean);   virtual;
    procedure Fuellen1; virtual;  // zufall
    procedure Fuellen2; virtual;  // teilweise
    procedure Fuellen3; virtual; // sortiert

    procedure Bubblesort1;                    virtual;
    procedure Bubblesort2;                    virtual;
    procedure Bubblesort3;                    virtual;
    procedure Shakersort1;                    virtual;
    procedure Auswahlsort;                    virtual;
    procedure quick;                          virtual;
    procedure Quicksort(anfang, ende: Cardinal; var act,sw : Cardinal);
    function  GetVergleiche :  Cardinal;       virtual;
    function  Getswaps :  Cardinal;            virtual;
  end;
(*  ----------------- B e s c h r e i b u n g -------------------------

Oberklasse     : -
Bezugsklassen  :
Methoden
--------
----------------------------------------------------------------------- *)


IMPLEMENTATION
(* ==================================================================== *)

var
  kunstwort : TWort;
  Vergleiche,
  swaps     : Cardinal;  // beim Sortieren

  Pegel     : TProgressbar;


procedure TWortListe.PegelAnzeigen(pb :   TProgressbar);
(* -------------------------------------------------------------------- *)
begin
  pegel := pb;
end;


procedure TWortListe.SetLen (L: Cardinal);
(* -------------------------------------------------------------------- *)
begin
  Listenlaenge := L;
end;

(* ====================== Liste fuellen =============================== *)
procedure TWortListe.PegelStart(max : Cardinal);
(* -------------------------------------------------------------------- *)
begin
  Pegel.Min := 0 ;
  Pegel.Position := 0;
  Pegel.Step := 1;
  Pegel.Max  := max;

  if   Pegel.Max >= 1000000
  then Pegel.Step := Pegel.Max DIV 10
  else Pegel.Step := 1;
end;


procedure TWortListe.SetElemLen (l: Cardinal);
(* -------------------------------------------------------------------- *)
begin
  ElemLaenge := l;
end;

procedure TWortListe.SetElemCase(c : boolean);
(* -------------------------------------------------------------------- *)
begin
  Elemcase := c;
end;

procedure TWortListe.SetSingleMode (m : boolean);
(* -------------------------------------------------------------------- *)
begin
  Singlemode := m;
end;

procedure TWortListe.NeuesWortMachen;
(* -------------------------------------------------------------------- *)
begin
  Kunstwort := TWort.Create;
  Kunstwort.SetLaenge(Elemlaenge);
  Kunstwort.SetUpCase(Elemcase);
end;

procedure TWortListe.Fuellen1; // zufall
(* -------------------------------------------------------------------- *)
var i : Cardinal;
begin
  NeuesWortMachen;
  Pegelstart(Listenlaenge);
  for i := 1 to Listenlaenge do
  begin
    Kunstwort.Erzeugen;
    Kollektion [i] := Kunstwort.GetWort;
    Pegel.Position := i;
  end;
end;

procedure TWortListe.Fuellen2; // letztes El. unsortiert
(* -------------------------------------------------------------------- *)
begin
  Fuellen1;
  Quicksort(1,(Listenlaenge - 1),Vergleiche,swaps);
end;

procedure TWortListe.Fuellen3; // sortiert
(* -------------------------------------------------------------------- *)
begin
  Fuellen1;
  Quicksort(1,(Listenlaenge),Vergleiche,swaps);
end;

(* ========================= Sortieren ================================ *)

procedure Tauschen (var E1, E2 : TElement);
(* ------------------------------------  *)
var
   hilf : TElement;
begin
    hilf := E1;
    E1  := E2;
    E2  := hilf
end;


procedure TWortListe.Bubblesort1;
(* -------------------------------------------------------------------- *)
(* Auftrag: lineares Sortieren durch systematisches Vergleichen aller   *)
(*          Elemente  und Tauschen mit dem nachfolgenden Element        *)
(* Vorher : Liste ist initialisiert                                     *)
(* Nachher: Listeist aufsteigend sortiert                               *)
(* -------------------------------------------------------------------- *)
var
  Durchlauf,
  i         : Cardinal;

begin
  if Listenlaenge >= 50000
  then PegelStart(sqr(Listenlaenge DIV 4))
  else PegelStart(sqr(Listenlaenge -1 ));

  Vergleiche := 0;
  Swaps      := 0;
  for Durchlauf := 1 to (Listenlaenge - 1) do
   begin
     for i := 1 to (Listenlaenge - 1) do
     begin
       if Kollektion[i] > Kollektion[i+1]
       then
        begin
          Tauschen(Kollektion[i], Kollektion[i+1]);
          inc(swaps);
        end;
        Vergleiche := Vergleiche +1;
     end;
     if Listenlaenge >= 50000
     then Pegel.Position := (Vergleiche DIV 1000)
     else Pegel.Position := Vergleiche;
   end; (* for Durchlauf *)
end;


procedure TWortListe.Bubblesort2;
(* -------------------------------------------------------------------  *)
(* Auftrag: lineares Sortieren durch Vertauschen des nachfolgenden      *)
(*          Elements im Restfeld, das von hinten verkleinert wird.      *)
(* Vorher : Liste ist initialisiert                                     *)
(* Nachher: Liste ist aufsteigend sortiert                              *)
(* -------------------------------------------------------------------- *)

var
  Durchlauf,
  i,
  Endmarke  : Cardinal;

begin
  Vergleiche := 0;
  Swaps := 0;
  if Listenlaenge >= 50000
  then PegelStart(round(0.5 * sqr(Listenlaenge DIV 4)))
  else PegelStart (round(0.5 * (sqr(Listenlaenge) - Listenlaenge)));
  Endmarke := Listenlaenge - 1;

  for Durchlauf := 1 to (Listenlaenge - 1) do
  begin
    for i := 1 to (Endmarke) do
    begin
       if Kollektion[i] > Kollektion[i+1]
       then
        begin
          Tauschen(Kollektion[i], Kollektion[i+1]);
          inc(swaps);
        end;
       inc(Vergleiche);
    end;
    Endmarke := Endmarke - 1;
    if Listenlaenge >= 50000
    then Pegel.Position := (Vergleiche DIV 1000)
    else Pegel.Position := Vergleiche;
  end; (* for Durchlauf *)
end;



procedure TWortListe.Bubblesort3;
(* ------------------------------------------------------------------ *)
(* Auftrag: lineares Sortieren durch Vergleichen und Vertauschen des  *)
(*          nachfolgenden Elements bis nicht mehr getauscht wird.     *)
(*          Endmarke rueckt vor und verkleinert Restfeld              *)
(* vorher : Kollektion ist initialisiert                              *)
(* nachher: Kollektion ist aufsteigend sortiert                       *)
(* ------------------------------------------------------------------ *)
var
  y,
  Endmarke : Cardinal;
  IstFertig   : boolean;

begin
  Vergleiche := 0;
  Swaps := 0;
  Endmarke  := Listenlaenge - 1 ;

  if Listenlaenge >= 50000
  then PegelStart(round(0.5 * sqr(Listenlaenge DIV 4)))
   else PegelStart (round(0.5 * (sqr(Listenlaenge) - Listenlaenge)));
  repeat
    IstFertig := true ;
    for y := 1 to endmarke do
    begin
      if Kollektion[y] > Kollektion[y+1]
      then
        begin
          Tauschen (Kollektion[y], Kollektion[y+1]);
          inc(swaps);
          IstFertig := false;
        end;
      inc(Vergleiche);
    end;
    Endmarke := Endmarke - 1;        (* Restfeld verkleinern *)
    if Listenlaenge >= 50000
    then Pegel.Position := (Vergleiche DIV 1000)
    else Pegel.Position := Vergleiche;
  until IstFertig;
end;

procedure TWortListe.shakersort1;
(* -------------------------------------------------------------------  *)
(*  Aufgabe:                                                            *)
(*  lineares Sortieren durch Vergleichen und Vertauschen des            *)
(*  nachfolgenden Elements.                                             *)
(*  Das Feld wird abwechselnd von links nach rechts durchlaufen,        *)
(*  dabei die Feldgrenzen nach innen verkleinert solange die linke      *)
(*  Grenze kleiner ist als die rechte.                                  *)
(* vorher : Kollektion ist initialisiert                                *)
(* nachher: Kollektion ist aufsteigend sortiert                         *)
(* -------------------------------------------------------------------  *)

var
    y,
    linkeGrenze,
    rechteGrenze,
    tauschPos     : Cardinal;


begin
   Vergleiche := 0;
   Swaps      := 0;
   if Listenlaenge >= 50000
   then PegelStart(round(0.33 * (sqr(Listenlaenge DIV 4))))
   else PegelStart(round(0.33 * (sqr(Listenlaenge -1 ))));


   linkeGrenze  := 1;
   rechteGrenze := Listenlaenge - 1;
   tauschPos    := linkeGrenze;

   while linkeGrenze < rechteGrenze do
   begin
      for y :=  linkeGrenze to rechteGrenze do
      begin
        if Kollektion[y] > Kollektion[y+1]
        then begin
           Tauschen(Kollektion[y], Kollektion[y+1]);
           swaps := swaps + 1;                   (* mitzaehlen *)
           tauschPos := y;
        end;
        inc(Vergleiche);

      end; (*   for y  *)
      rechteGrenze := tauschPos - 1;

      for y :=  rechteGrenze downto linkeGrenze  do
      begin
        if Kollektion[y] > Kollektion[y+1]
        then begin
           Tauschen(Kollektion[y], Kollektion[y+1]);
           swaps := swaps + 1;                   (* mitzaehlen *)
           tauschPos := y;
        end;
       inc(Vergleiche);              (* mitzaehlen *)
      end; (*   for y  *)
      linkeGrenze := tauschPos +1;

      if Listenlaenge >= 50000
      then Pegel.Position := (Vergleiche DIV 10000)
      else Pegel.Position := Vergleiche;
   end; (*   while  *)
end;

procedure TWortListe.auswahlsort;
(* -------------------------------------------------------------------  *)
(* Aufgabe:  lineares Sortieren durch direkte Auswahl und Vertauschen   *)
(*           der Position.                                              *)
(* vorher : Kollektion ist initialisiert                                *)
(* nachher: Kollektion ist aufsteigend sortiert                         *)
(* -------------------------------------------------------------------  *)

var
   i,
   y,
   indexmin : Cardinal;                  (* Index des kleinsten Elem. *)

begin
  Vergleiche := 0;
  Swaps := 0;
  if Listenlaenge >= 50000
  then PegelStart(round(0.5 * (sqr(Listenlaenge DIV 4))))
  else PegelStart(round(0.5 * (sqr(Listenlaenge -1 ))));

   for i := 1 to (Listenlaenge - 1) do
   begin
      indexmin  := i;

      for y := (i + 1) to Listenlaenge do
      begin                          (*   im Restfeld das kleinste *)
        if Kollektion[y] < Kollektion[indexmin]      (*   Element suchen  *)
        then
         begin
           indexmin := y;
           inc(swaps);
         end;
         inc(Vergleiche);
      end; (* for y *)

      Tauschen(Kollektion[i], Kollektion[indexmin]);     (*   Positionen tauschen  *)

      if Listenlaenge >= 50000
      then Pegel.Position := (Vergleiche DIV 1000)
      else Pegel.Position := Vergleiche;
   end; (* for i *)
end;



procedure TWortListe.quick;
(* -------------------------------------------------------------------- *)
begin
  Vergleiche := 0;
  Swaps      := 0;
  PegelStart(round(Listenlaenge * log2(Listenlaenge)));
  quicksort(1, Listenlaenge, vergleiche,swaps);
end;

procedure TWortListe.quicksort(anfang, ende: Cardinal; var act,sw : Cardinal);
(* -------------------------------------------------------------------- *)
var linkerzeiger,
    rechterzeiger : Cardinal;
    mitte         : string;

begin

  linkerzeiger  := anfang;
  rechterzeiger := ende;
  mitte         := Kollektion[(linkerzeiger + rechterzeiger) div 2];

  Pegel.Position := Vergleiche ;
  repeat
   while (Kollektion[linkerzeiger] < mitte) do
   begin
    linkerzeiger := linkerzeiger + 1;
    inc(act);
   end;

   while (Kollektion[rechterzeiger] > mitte) do
   begin
     rechterzeiger := rechterzeiger-1;
     inc(act);
   end;

   if linkerzeiger <= rechterzeiger
   then
     begin
       Tauschen(Kollektion[linkerzeiger], Kollektion[rechterzeiger]);
       linkerzeiger := linkerzeiger + 1;
       rechterzeiger := rechterzeiger - 1;
       inc(Swaps);
       inc(act);
     end;  (* if *)
  until rechterzeiger < linkerzeiger;


  if anfang < rechterzeiger
  then quicksort(anfang, rechterzeiger,Vergleiche,swaps);

  if linkerzeiger < ende
  then quicksort(linkerzeiger, ende,Vergleiche, swaps);
end; (* quicksort *)


function  TWortListe.GetVergleiche :  Cardinal;
(* -------------------------------------------------------------------- *)
begin
   Result := Vergleiche;
end;

function  TWortListe.GetSwaps :  Cardinal;
(* -------------------------------------------------------------------- *)
begin
   Result := Swaps;
end;

END.
